      Program VGS
C     ***************************************************************                          
C     Estimation of soil hydraulic parameters. 
C     Model: van Genuchten/Mualem or Soto Fractal model
C
C     unimodal, option:simultaneous fit to water retention (WRC)
C                      AND hydraulic conductivity function (HCF)
C     vanGenuchten (1980): SSSAJ 44:892-898
C     Soto,M.A.,H.K.Chang, M.Th.van Genuchten(2017): Fractal-basesd 
C     models for the unsaturated soil hydraulic functions.-
C     Geoderma 306:144-151
C
C     1st step: Find global error minimum (WRC only) with Monte Carlo
C     2nd step: Fibonacci algorithm (WRC and HCF)
C     ****************************************************************

      IMPLICIT REAL*8 (A-H,O-Z)
      integer   soilID
      dimension RMSE(3),ran(4,90000),hp(20)   
      character*8  inputname
      character*6   ID
      Common /RN/   ZR1(90000),ZR2(90000),ZR3(90000),ZR4(90000)  
      COMMON /FIBOZ/FB(0:50)              ! Fibonacci numbers 
      common /par/  p(6),bott(6),top(6),ha   
      COMMON /ARRAY/hpR(20),hpC(20),th(20), WC(20),condobs(20),
     # condobsL(20), cond(20), w, ksimul,model,kMCM,khc 
      external OBJF,VANGEN, SOTO

C     -----calculation of Fibonacci numbers ----
      FB(0)=1;FB(1)=1;
      DO 2  i=2,50
    2 FB(i)=FB(i-2)+FB(i-1)

      X=1
      call RANDOM_NUMBER (Harvest=X)   ! Random field generation
      call RANDOM_NUMBER (ran)
C
      print *,'    '
      print *,'  Program VGS to estimate the parameters ',
     # 'of the vanGenuchten or the Soto model in two steps'
      print *,'    '
 
      print *,'  enter name of input file, max 8 characters '
      read (*,*) inputname
      OPEN (17,  file=inputname,  STATUS='old')            ! indata
      open (18,  file='Out'//inputname  ,status='unknown') ! out Parameter
      open (19,  file='HyCond',   status='unknown')        ! output hy.cond.

      write (18,700)
  700 format(//,5X,' *****   Program VGS     *****',/,
     # 10X,'    ==================')
      
      print *,'  For vanGenuchten enter 1, for Soto enter 2'
      read (*,*)     model
      write (18,702) model
  702 format(2X,/,5X,' model= ',I2,/,2X)

      print *,' for theta-fit enter 0,for simultaneous fit enter 1'
      read  (*,*)    ksimul
      write (18,703) ksimul
  703 format(5X,'KSIMUL= ',i2,/,2X)
C
      Nsoils=0
C     +++++++++    START     +++++++++++++++++++++++++++
  100 continue    ! return point: next soil

      nwrc=0     ! number of water retention data points
      ncond=0    ! number of hy. conductivity data
      soilID=0
      RMSE(3)=0.
   
      do i=1,20
       cond(i)=0.    ! calculated relative conductivity, logarithmic
       condobs(i)=0. ! observed
       condobsL(i)=0.! obs., logarithmic
       th(i)=0. ! observed
       WC(i)=0. ! calculated
      end do
C
C     conductivity:  cond calculated, condobs observed, ncond data sets
C     water content: th observed, WC calculated, nwrc data sets
C     RMSE(1): fit of WRC using MCM (VANGEN or SOTO)
C     RMSE(2): fit of WRC using Fibonacci (VANGEN or SOTO)
C     RMSE(3): fit of HCF using the Fibonacci algorithm
	
C     -----  data input      -----
      if (ksimul == 0) read (17,*,END=500) soilID, nwrc, conds
      ncond=nwrc  ! gets overridden in case of ksimul=1
C     saturated hy. cond.(conds) needed for prediction purposes
C     chose conds=1 for output of relative conductivity

      if (ksimul == 1) read (17,*,End=500) soilID, nwrc,ncond, conds
C     ksimul=0:To predict hydraulic conductivity,an estimated conds must be given
C     ksimul=1:The conds from the input file is regarded as initial estimate

      Nsoils=Nsoils+1
      if (nwrc > 20 .or. ncond > 20) then
         print *,soilID,
     # ' Number of data points of each kind must not exceed 20 '
         goto 500
      end if

      read (17,*) (hpR(j),j=1,nwrc) ! read pressure head data
      read (17,*) (th(j), j=1,nwrc) ! read observed water content
      do j=1,nwrc
       if (th(j) >1.) th(j)=0.01*th(j) ! vol% to cm3/cm3
       hpC(j)=hpR(j)  ! replaced later if necessary
      end do
         
C     -----     simultaneous fit   -----
C     hpC pressure head data corresponding to hydraulic conductivity 
      if (ksimul.eq. 1) then
        read (17,*) (hpC(j),    j=1,ncond) 
        read (17,*) (condobs(j),j=1,ncond) !read hydraulic conductivity 
        sumth=0.
        sumcond=0.
        do j=1, ncond
          condobsL(j)=DLOG10(condobs(j))    ! log values of observed conductivity
          sumcond=sumcond+DABS(condobsL(j))  
        end do
C      -----     prepare weight     -----
       do i=1,nwrc
         sumth=sumth+th(i)  
       end do
       w=(sumth/nwrc)/(sumcond/ncond) 
C       print *,'w= ',w
      end if  ! end of case ksimul=1
C     end of data input

C     ------------------------------
C     p   vanGenuchten   Soto et al.
C     ------------------------------
C     1   thr            thr
C     2   ths            ths
C     3   alpha          D
C     4   vgn            hzero
C     5   tau            ---
C     6   conds          conds
C     ------------------------------ 
      
C     ---   parameter bounds   --- 
      npar=4
C     -----  van Genuchten and Soto model ---
      bott(1)=0.01  ! theta-r
      top (1)=0.99*th(nwrc) ! decreasing theta assumed
      bott(2)=th(1) ! theta-s
      top (2)=0.65
C     ------  Soto model    -------
      if (model == 2) then
        bott(3)=1.1    ! D
        top (3)=2.98   
        bott(4)=1.     ! hzero
        top (4)=1.0D+06   !
      else
C     ------ van Genuchten model  ------
        bott(3)=0.005 ! vanGen-alpha
        top (3)=0.7
        bott(4)=1.05  ! vanGen-n
        top (4)=8.
      end if

C     Monte-Carlo search for global minimum of WRC errors -
C     ran_i is an array of equally distributed random numbers between
C     zero and unity.

C     Re-scale random numbers according to given bounds     
      Nrandom=90000
      do j=1, Nrandom
       ZR1(j)=bott(1)+(top(1)-bott(1))*ran(1,j)
       ZR2(j)=bott(2)+(top(2)-bott(2))*ran(2,j)
       ZR3(j)=bott(3)+(top(3)-bott(3))*ran(3,j)
       ZR4(j)=bott(4)+(top(4)-bott(4))*ran(4,j)
      end do
      if (model == 1) then  ! van Genuchten
        call MCM (Nrandom,nwrc,ncond,VANGEN,RMSE(1))
      else   ! fractal model Soto et al.
        call MCM (Nrandom,nwrc,ncond,SOTO,RMSE(1)) 
      end if
      kMCM=0   ! in MCM kMCM was set to 1
C
C     -----    Step 2: Fibonacci parameter optimization     -----
C     wrc - parameter bounds vanGenuchten model as above (full range)
C     fractal model:narrow range for FIBONACCI search because
c     the 4D-mountains of variances are very rugged
      if (model ==2) then 
       do i=1,4
       bott(i)=0.8*p(i)
       top (i)=1.2*p(i)
       end do
      end if 
      npar=4
C     prediction of hy. cond. without simultaneous fit:
      p(5)=0.5     ! tortuosity standard value
      p(6)=conds   ! from input, used for predicition or as initial estimate
C
      if (ksimul == 1) then !simultaneous fit 
        npar=6
        bott(5)=-5.  ! tortuosity range, not used in Soto
        top (5)= 3.
        bott(6)= 0.1*conds
        top (6)= 10.*conds
      end if 
C     explanation: while Fibonacci is looking for p(i), all other parameters
C     keep their previous values which were obtained either by MCM or by 
C     the previous Fibonacci iteration or are set as above 

      TRMSE=100.; niter=0; NF=30;
      if (model == 1) eps=5.0D-07
      if (model == 2) eps=5.0D-08

      do 120 ITER=1,500  ! Iterations loop 
       prE=TRMSE         ! previous RMSE total
       do 125 jp=1,npar   
C      -----   Fibonacci called for each parameter   -----
        if (model ==1) then
         call FIBO (NF,VANGEN,nwrc,ncond,jp,model,TRMSE)
        else
         call FIBO (NF,SOTO,  nwrc,ncond,jp,model,TRMSE)
        end if
  125  continue  ! all the parameters are fitted once   
       niter=niter+1
       if (DABS(TRMSE-prE) .le. eps) goto 130
  120 continue 
  130 continue
      print *,'soil,niter ',nsoils,niter

C      ---- RMSE(2) of water retention for all options   ----      
      S1=0.
      do i=1, nwrc     ! number of wrc-points
         D=th(i)-WC(i) ! observed th,  calculated wc
         S1=S1+D*D
      end do
      RMSE(2)=DSQRT(S1/dfloat(nwrc))  
C     
      if (ksimul == 1) then  
        sumq=0.
        do i=1, ncond
        sumq=sumq+(condobsL(i)-cond(i))**2
        end do
        RMSE(3)=DSQRT(sumq/float(ncond-1))
C       the weighted RMSE of simultaneous fitting is not shown
      end if
C
C     ----- output     -----
C     output into file 18
      if (model ==  2) p(5)=ha
      if (model ==1 .and. ksimul == 0) p(5)=0.
      if (model ==1 ) write (18,705) 
      if (model ==2 ) write (18,706) 

  705 format(//,2X,'SOIL model sim   theta-r       theta-s       ',
     # 'alpha       vgn',9X,'tortuosity   conds          ',
     # 'RMSE 1      RMSE2       RMSE 3 ')
  706 format(//,2X,'Soil model sim  theta-r      theta-s           D',
     # '          h-zero        ha        conds     ',
     # '     RMSE 1       RMSE 2       RMSE 3')
      write (18,710) soilID,model,ksimul,(p(j),j=1,6),(RMSE(j),j=1,3)       
  710 format (2X,3(I4,1X),9(E12.5,1X))     
      
C     output into file 19
      write (19,708)
  708 format(2X,'soil  model    ksim  ')   
      write (19,712) soilID,model,ksimul 
  712 format(2x, 4(I4,3X))
C
      write (19,714)  ! heading
  714 format(2X,' i        hpR       th(obs)       WC ')  
      do i=1,nwrc
      write (19, 716) i,hpR(i),th(i),wc(i)
  716 format(2X,I2,2X,3(E12.5,2X))
      end do
      write(19,718) ! heading
  718 format(/,2X,' i        hpC        condobs    condobsLog  ',
     #  'log10(cond) ') 
      do i=1,ncond
       write (19, 720) i,hpC(i),condobs(i),condobsL(i),cond(i)
  720 format(2X,I2,2X,4(E12.5,1X))
      end do
  
      goto 100                  
C
C     -----     finish     -----
  500 continue
      write (18,777) Nsoils
  777 format(//,2X, I3,2X,'soils')
      print *,' done '
      stop
      end   
 

      SUBROUTINE MCM (Nrandom,nth,nobk,FUNCT,RMSD) 
C     ************************************************
      IMPLICIT REAL*8 (A-H,O-Z)
      dimension best(4)
      Common /RN/   ZR1(90000),ZR2(90000),ZR3(90000),ZR4(90000)
      common /par/  p(6),bott(6),top(6),ha    
      COMMON /ARRAY/hpR(20),hpC(20),th(20), WC(20),condobs(20),
     # condobsL(20), cond(20), w, ksimul,model,kMCM,khc 
C
      kMCM=1
      FMIN=90000.

      do 100 j=1,Nrandom 
       p(1) =ZR1(j); p(2) =ZR2(j); p(3) =ZR3(j); p(4) =ZR4(j);
       call FUNCT (nth,nobk)
       call OBJF  (nth,nobk, RS)
      if (RS .lt. FMIN) then
        FMIN=RS
        do 25 k=1,4
   25   best(k)=p(k) 
      end if
  100 continue
      do 30 k=1,4
   30 p(k)=best(k)
      RMSD=FMIN
      RETURN
      END

      SUBROUTINE OBJF (nth,nobk, RS)
C     *********************************
C     ---- PURPOSE: TO CALCULATE THE GOODNESS OF FIT ----
      implicit real*8 (A-H,O-Z)   
      COMMON /ARRAY/hpR(20),hpC(20),th(20), WC(20),condobs(20),
     # condobsL(20), cond(20), w, ksimul,model,kMCM,khc 

      S1=0.0; S2=0.
      DO i=1, nth   ! number of wrc-points
         D=th(i)-WC(i) ! observed th,  calculated wc
         S1=S1+D*D
      end do
      RS=DSQRT(S1/dfloat(nth))
      if (kMCM == 1 .or. ksimul == 0) return
C
      do i=1,nobk       
        D=w*(condobsL(i)-cond(i))
        S2=S2+D*D
      end do
      RS=DSQRT((S1+S2)/DFLOAT(nth+nobk-1))
      RETURN
      END

      SUBROUTINE VANGEN (nth,nobk)
C     ****************************
C     -----    purpose: To calculate functional values     -----
      implicit real*8 (A-H,O-Z)
      common /par/  p(6),bott(6),top(6),ha  
      COMMON /ARRAY/hpR(20),hpC(20),th(20), WC(20),condobs(20),
     # condobsL(20), cond(20), w, ksimul,model,kMCM,khc 
    
      vgm=1.-1./p(4)
      c1=p(3)**p(4)   ! vga**vgn
      c2=p(2)-p(1)    ! ths-thr

C     ---   calculate water content   ---    
     
      do 10 i=1,nth
       c3=hpR(i)**p(4)  ! hp **vgn
       c4=1.+c1*c3    
       c5=c4**vgm    ! denumerator
       WC(i)=p(1)+c2/c5
   10 continue  
      if (kMCM == 1) return
C      ---   calculate hydraulic conductivity ---
      p(5)=DMAX1(p(5),1.-2./vgm)  ! constraint Peters et al. (2011) 
      do i=1,nobk 
       c3=hpC(i)**p(4)  ! hp **vgn
       c4=1.+c1*c3    
       c5=c4**vgm 
       c6=1.-c1*c3/(p(3)*hpC(i)*c5)
       cond(i)=dlog10(p(6)*(c6*c6/c5**p(5))) ! log of relative hydraulic conductivity
      end do
      return
      end

      Subroutine SOTO (nth,nobk)
C     **************************
C     Alfaro Soto, M.A., Chang, H.K., vanGenuchten, M.Th:
C     Fractal-based models for the unsaturated soil hydraulic
C     functions.- Geoderma 306 (2017) 144-151

      implicit real*8 (A-H,O-Z)
      common /par/  p(6),bott(6),top(6),ha  
      COMMON /ARRAY/hpR(20),hpC(20),th(20), WC(20),condobs(20),
     # condobsL(20), cond(20), w, ksimul,model,kMCM,khc 

      thr=p(1); ths=p(2); D=p(3); plambda=3.-D; hzero=p(4);
      Sr=thr/ths
      ha=hzero*(Sr/(1.-Sr))**(1./plambda)

      do  i=1, nth
        WC(i)=ths
        if (hpR(i) > ha) WC(i)=thr*(1.+(hpR(i)/hzero)**(-plambda))
      end do
      if (kMCM ==1 ) return

      do i=1, nobk
        wcc=ths
        if (hpC(i) > ha) wcc=thr*(1.+(hpC(i)/hzero)**(-plambda))    
        Se=(wcc-thr)/(ths-thr)
        cond(i)=dlog10(p(6)*(Se**((19.-5.*D)/(6.-2.*D)))) 
      end do
      return
      end


      SUBROUTINE Fibo(NF,FUNCT,nth,nobk,jp,model,RMSD)
C     ************************************************ 
      implicit real*8 (A-H,O-Z)
      real*8 LB  ! lower bound
      common /FIBOZ/ FB(0:50)
      common /par/  p(6),bott(6),top(6),ha                
C
      if (model == 2 .and. jp == 5) return ! p(5) not used by Soto
      LB=bott(jp)  ! low bound
      HB=top (jp)  ! high bound

      D=(HB-LB)*FB(NF-2)/FB(NF)
      X1=LB+D; X2=HB-D;p(jp)=X1;
      CALL FUNCT(nth,nobk)
      CALL OBJF (nth,nobk,f1)
      p(jp)=X2
      CALL FUNCT(nth,nobk)
      CALL OBJF (nth,nobk,f2)

      do i=2, NF-1
       d=x2-x1
       if(i == (NF-1)) D=1.1*D

       if (f1 > f2)  then
        LB=X1; x1=HB-D;p(jp)=X1;
        call FUNCT(nth,nobk); call OBJF (nth,nobk,f1)
       else
        HB=X2;X2=LB+D;p(jp)=X2;
        call FUNCT(nth,nobk)
        call OBJF (nth,nobk,f2)
       endif

       if(X1 > X2) then
        hv=X1;X1=X2;X2=hv;hv=f1;f1=f2;f2=hv;
       end if
      end do 

      xm=.5*(HB+LB)
      p(jp)=xm
      call FUNCT(nth,nobk)
      call OBJF (nth,nobk,RS)
      RMSD=RS
      return
      end
      

